# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['volttron',
 'volttron.client',
 'volttron.client.commands',
 'volttron.client.messaging',
 'volttron.client.vip',
 'volttron.client.vip.agent',
 'volttron.client.vip.agent.subsystems',
 'volttron.server',
 'volttron.server.router',
 'volttron.services.auth',
 'volttron.services.config_store',
 'volttron.services.control',
 'volttron.services.health',
 'volttron.services.routing',
 'volttron.types',
 'volttron.utils']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'cryptography>=36.0.1,<37.0.0',
 'dateutils>=0.6.12,<0.7.0',
 'gevent>=21.12.0,<22.0.0',
 'pip==22.2.2',
 'psutil>=5.9.0,<6.0.0',
 'pyzmq>=22.3.0,<23.0.0',
 'toml>=0.10.2,<0.11.0',
 'tzlocal>=4.1,<5.0',
 'watchdog-gevent>=0.1.1,<0.2.0']

entry_points = \
{'console_scripts': ['vcfg = volttron.client.commands.config:main',
                     'vctl = volttron.client.commands.control:main',
                     'volttron = volttron.server.__main__:main',
                     'volttron-cfg = volttron.client.commands.config:main',
                     'volttron-ctl = volttron.client.commands.control:main']}

setup_kwargs = {
    'name': 'volttron',
    'version': '10.0.2rc0',
    'description': 'VOLTTRON™ is an open source platform for distributed sensing and control. The platform provides services for collecting and storing data from buildings and devices and provides an environment for developing applications which interact with that data.',
    'long_description': 'VOLTTRON™ is an open source platform for distributed sensing and control. The platform provides services for collecting and storing data from buildings and devices and provides an environment for developing applications which interact with that data.\n\n[![Pytests](https://github.com/eclipse-volttron/volttron-core/actions/workflows/run-tests.yml/badge.svg)](https://github.com/eclipse-volttron/volttron-core/actions/workflows/run-tests.yml)\n[![pypi version](https://img.shields.io/pypi/v/volttron.svg)](https://pypi.org/project/volttron/)\n\n## Installation\n\nIt is recommended to use a virtual environment for installing volttron.\n\n```shell\npython -m venv env\nsource env/bin/activate\n\npip install volttron\n```\n\n### Quick Start\n\n 1. Setup VOLTTRON_HOME environment variable: export VOLTTRON_HOME=/path/to/volttron_home/dir \n \n    **NOTE** This is madatory if you have/had in the past, a monolithic    VOLTTRON version that used the default VOLTTRON_HOME $HOME/.volttron. This modular version of VOLTTRON cannot work with volttron_home used by monolithic version of VOLTTRON(version 8.3 or earlier)\n \n 2. Start the platform\n    ```bash\n    > volttron -vv -l volttron.log &>/dev/null &\n    ```\n\n 3. Install listener agent\n    ```bash\n    > vctl install volttron-listener --start\n    ```\n\n 4. View status of platform\n    ```bash\n    > vctl status\n    ```\n\n 5. Shutdown the platform\n    ```bash\n    > vctl shutdown --platform\n    ```\n\nFull VOLTTRON documentation available at [VOLTTRON Readthedocs](https://volttron.readthedocs.io)\n\n## Contributing to VOLTTRON\n\nPlease see the [contributing.md](CONTRIBUTING.md) document before contributing to this repository.\n\nPlease see [developing_on_modular.md](DEVELOPING_ON_MODULAR.md) document for developing your agents against volttron.\n',
    'author': 'volttron',
    'author_email': 'volttron@pnnl.gov',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://volttron.org',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

"""VOLTTRON platform™ messaging classes."""

import zmq
from ...utils import jsonapi

from .headers import Headers

__all__ = ["Headers", "Socket"]

__author__ = "Brandon Carpenter <brandon.carpenter@pnnl.gov>"
__copyright__ = "Copyright (c) 2016, Battelle Memorial Institute"
__license__ = "Apache 2.0"


class Socket(zmq.Socket):
    """ØMQ socket with additional agent messaging methods."""

    def __new__(cls, socket_type, context=None):
        if not context:
            context = zmq.Context.instance()
        return zmq.Socket.__new__(cls, context, socket_type)

    def __init__(self, socket_type, context=None):
        super(Socket, self).__init__(self.context, socket_type)

    # Override send_string to ensure copy defaults to True.
    # https://github.com/zeromq/pyzmq/pull/456
    def send_string(self, u, flags=0, copy=True, encoding="utf-8"):
        super(Socket, self).send_string(u, flags=flags, copy=copy, encoding=encoding)

    send_string.__doc__ = zmq.Socket.send_string.__doc__

    def recv_message(self, flags=0):
        """Recieve a message as (topic, headers, message) tuple."""
        topic = self.recv_string(flags)
        headers = self.recv_string(flags) if self.rcvmore else ""
        headers = jsonapi.loads(headers) if headers else {}
        message = self.recv_multipart(flags) if self.rcvmore else []
        return topic, Headers(headers), message

    def recv_message_ex(self, flags=0):
        """Receive a message as (content type, message) tuples.

        Like recv_message(), returns a three tuple.  However, the final
        message component contains a list of 2-tuples instead of a list
        of messages.  These 2-tuples contain the content- type and the
        data.
        """
        topic, headers, message = self.recv_message(flags)
        message = list(zip(headers["Content-Type"], message))
        return topic, headers, message

    def send_message(self, topic, headers, *msg_parts, **kwargs):
        """Send a multipart message with topic and headers.

        Send a multipart message on the socket with topic being a UTF-8
        string, headers can be a dictionary or a Headers object, and
        msg_parts is the optional parts of the message.  The media or
        content type of each message component should be included in the
        'Content-Type' header which should be a list of MIME types or a
        string if there is only one message part.
        """
        flags = kwargs.pop("flags", 0)
        if kwargs:
            raise TypeError("send_message() got unexpected keyword "
                            "arugment(s): " + ", ".join(kwargs))
        if not isinstance(headers, Headers):
            headers = Headers(headers) if headers else Headers()
        self.send_string(topic, flags | zmq.SNDMORE)
        self.send_json(headers.dict, flags | (zmq.SNDMORE if msg_parts else 0))
        if msg_parts:
            self.send_multipart(msg_parts, flags)

    def send_message_ex(self, topic, headers, *msg_tuples, **kwargs):
        """Send messages given as (content-type, message) tuples.

        Similar to the send_message method except that messages are given as
        2-tuples with the MIME type as the first element and the message
        data as the second element.
        """
        headers = Headers(headers) if headers else Headers()
        headers["Content-Type"], msg_parts = list(zip(*msg_tuples))
        self.send_message(topic, headers.dict, *msg_parts, **kwargs)



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

"""Utilities for tracking VIP message statistics at the router."""

import gevent

from volttron.server.router import UNROUTABLE, ERROR, INCOMING

__all__ = ["Tracker"]


def pick(frames, index):
    """Return the frame at index, converted to bytes, or None."""
    try:
        return bytes(frames[index])
    except IndexError:
        return None


def increment(prop, key):
    """Increment or set to 1 the value in prop[key]."""
    try:
        prop[key] += 1
    except KeyError:
        prop[key] = 1


class Tracker(object):
    """Object for sharing data between the router and control objects."""

    def __init__(self):
        self._reset()
        self.enabled = False

    def reset(self):
        """Reset all counters to default values and set start time."""
        self._reset()
        self.stats["start"] = gevent.get_hub().loop.now()

    def _reset(self):
        """Initialize statistics counters."""
        self.stats = {
            "error": {
                "error": {},
                "peer": {},
                "user": {},
                "subsystem": {}
            },
            "unroutable": {
                "error": {},
                "peer": {}
            },
            "incoming": {
                "peer": {},
                "user": {},
                "subsystem": {}
            },
            "outgoing": {
                "peer": {},
                "user": {},
                "subsystem": {}
            },
        }

    def hit(self, topic, frames, extra):
        """Increment counters for given topic and frames."""
        if self.enabled:
            if topic == UNROUTABLE:
                stat = self.stats["unroutable"]
                increment(stat["error"], extra)
            else:
                user = pick(frames, 3)
                subsystem = pick(frames, 5)
                if topic == ERROR:
                    stat = self.stats["error"]
                    increment(stat["error"], bytes(extra[0]))
                else:
                    stat = self.stats["incoming" if topic == INCOMING else "outgoing"]
                increment(stat["user"], user)
                increment(stat["subsystem"], subsystem)
            increment(stat["peer"], pick(frames, 0))

    def enable(self):
        """Enable tracking."""
        if not self.enabled:
            self.reset()
            self.enabled = True

    def disable(self):
        """Disable tracking."""
        if self.enabled:
            self.enabled = False
            self.stats["end"] = gevent.get_hub().loop.now()



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

import weakref

__all__ = ["Signal"]


class Signal(object):

    def __init__(self):
        self._receivers = weakref.WeakValueDictionary()

    def connect(self, receiver, owner=None):
        self._receivers[receiver] = receiver if owner is None else owner

    def disconnect(self, receiver):
        try:
            self._receivers.pop(receiver)
            return True
        except KeyError:
            return False

    def send(self, sender, **kwargs):
        return [receiver(sender, **kwargs) for receiver in self._receivers]

    def sendby(self, executor, sender, **kwargs):
        return [executor(receiver, sender, **kwargs) for receiver in self._receivers]

    def receiver(self, func):
        self.connect(func)
        return func

    def __bool__(self):
        return bool(self._receivers)



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

""" Core package."""
from typing import List

from gevent import monkey

# At this point these are the only things that need to be patched
# and the server and client are working harmoniously with this.
patches = [
    ('ssl', monkey.patch_ssl),
    ('socket', monkey.patch_socket),
    ('os', monkey.patch_os),
]

# patch modules if necessary.  Only if the module hasn't been patched before.
# this could happen if the server code uses the client (which it does).
for module, fn in patches:
    if not monkey.is_module_patched(module):
        fn()

import logging

from urllib.parse import urlparse

from volttron.client.vip.agent import Agent, build_agent
from volttron.client.vip.agent.core import Core
from volttron.client.vip.agent.subsystems.rpc import RPC
from volttron.client.vip.agent.subsystems.channel import Channel
from volttron.client.vip.agent.subsystems.hello import Hello
from volttron.client.vip.agent.subsystems.peerlist import PeerList
from volttron.client.vip.agent.subsystems.ping import Ping
from volttron.client.vip.agent.subsystems.pubsub import PubSub
from volttron.client.vip.agent.subsystems.rpc import RPC
from volttron.client.vip.agent.subsystems.heartbeat import Heartbeat
from volttron.client.vip.agent.subsystems.health import Health
from volttron.client.vip.agent.subsystems.configstore import ConfigStore
from volttron.client.vip.agent.subsystems.auth import Auth
from volttron.client.vip.agent.subsystems.query import Query

__all__: List[str] = [
    "Agent", "Core", "RPC", "Channel", "Hello", "PeerList", "Ping", "PubSub", "Heartbeat",
    "Health", "ConfigStore", "Auth", "Query"
]

_log = logging.getLogger(__name__)


def build_vip_address_string(vip_root, serverkey, publickey, secretkey):
    """Build a full vip address string based upon the passed arguments

    All arguments are required to be non-None in order for the string to be
    created successfully.

    :raises ValueError if one of the parameters is None.
    """
    _log.debug("root: {}, serverkey: {}, publickey: {}, secretkey: {}".format(
        vip_root, serverkey, publickey, secretkey))
    parsed = urlparse(vip_root)
    if parsed.scheme == "tcp":
        if not (serverkey and publickey and secretkey and vip_root):
            raise ValueError("All parameters must be entered.")

        root = "{}?serverkey={}&publickey={}&secretkey={}".format(vip_root, serverkey, publickey,
                                                                  secretkey)

    elif parsed.scheme == "ipc":
        root = vip_root
    else:
        raise ValueError("Invalid vip root specified!")

    return root


# def update_volttron_script_path(path: str) -> str:
#     """
#     Assumes that path's current working directory is in the root directory of the volttron codebase.

#     Prepend 'VOLTTRON_ROOT' to internal volttron script if 'VOLTTRON_ROOT' is set and return new path;
#     otherwise, return original path
#     :param path: relative path to the internal volttron script
#     :return: updated path to volttron script
#     """
#     if os.environ["VOLTTRON_ROOT"]:
#         args = path.split("/")
#         path = f"{os.path.join(os.environ['VOLTTRON_ROOT'], *args)}"
#     _log.debug(f"Path to script: {path}")
#     return path



# -*- coding: utf-8 -*- {{{
# ===----------------------------------------------------------------------===
#
#                 Installable Component of Eclipse VOLTTRON
#
# ===----------------------------------------------------------------------===
#
# Copyright 2022 Battelle Memorial Institute
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# ===----------------------------------------------------------------------===
# }}}

from __future__ import annotations

import importlib
import logging
import pkgutil
from copy import copy
from types import ModuleType
from typing import List, Tuple, Dict, Set, KeysView

from volttron.types.server_config import ServerConfig
from volttron.types import ServiceInterface
from volttron.utils import get_subclasses, get_module, get_class

_log = logging.getLogger(__name__)

__discovered_plugins__: Dict[str, Tuple] = {}
__namespaces__: Set[str] = set()
__required_plugins__: Set[str] = set()
__plugin_startup_order__: List[str] = [
    "volttron.services.config_store",
    "volttron.services.auth",
]
__disabled_plugins__: Set[str] = set()

__all__ = ["get_service_names", "init_services", "discover_services"]

__service_interface_class__ = get_class('volttron.types', 'ServiceInterface')


def get_service_names() -> KeysView[str]:
    return __discovered_plugins__.keys()


def start_service(service_name: str):
    service_interface = get_class('volttron.types', 'ServiceInterface')
    module = get_module(service_name)
    subclasses = get_subclasses(module, service_interface)
    service = subclasses[0](serverkey=None,
                            identity="platform.web",
                            bind_web_address="http://127.0.0.1:8080")
    greenlet = service.spawn_in_greenlet()
    return greenlet


def init_services(config: ServerConfig) -> List[Tuple[str, ServiceInterface]]:

    default_kwargs = dict(enable_store=False,
                          address=config.internal_address,
                          heartbeat_autostart=True)

    found = []
    if not __disabled_plugins__:
        found = __discover_services__("volttron.services")

    def init_plugin(plugin_name, plugin):
        try:
            kwargs = copy(default_kwargs)
            module = plugin    # get_module("volttron.services")
            subclasses = get_subclasses(module, __service_interface_class__, return_all=True)
            kwargs.update(config.get_service_kwargs(plugin_name))
            if "identity" not in kwargs:
                kwargs["identity"] = plugin_name.replace("volttron.services", "platform")
            # TODO allow multiple subclasses in file suppport.
            for sub in subclasses:
                # TODO replace try except block with inspection of parameters
                try:
                    instance = sub(server_config=config, **kwargs)
                except TypeError:
                    _log.debug(f"{plugin_name} does not take a server_config as first param.")
                    instance = sub(**kwargs)

                return instance
        except ValueError as ex:
            _log.warning(ex.args)

    inited_services = []

    for plugin_name in __plugin_startup_order__:
        if plugin_name not in found:
            raise ValueError(f"Invalid plugin specified in plugin_startup_order {plugin_name}")
        _log.info(f"Init plugin: {plugin_name}, {__discovered_plugins__[plugin_name]}")
        if plugin_name not in __disabled_plugins__:
            plugin = __discovered_plugins__.pop(plugin_name)
            inited_services.append((plugin_name, init_plugin(plugin_name, plugin)))

    for plugin_name, plugin in __discovered_plugins__.items():
        if plugin_name not in __plugin_startup_order__ and plugin_name not in __disabled_plugins__:
            if config.get_service_enabled(plugin_name):
                _log.info(f"Init plugin {plugin_name}, {plugin}")
                service = init_plugin(plugin_name, plugin)
                # Only add service to the list if the service has the correct interface.
                if service is not None:
                    inited_services.append((plugin_name, service))
            else:
                _log.debug(f"Plugin {plugin_name} not enabled.")

    return inited_services


def discover_services(namespace: str) -> List[str]:
    module = importlib.import_module(namespace)
    return __discover_services__(module)


def __iter_namespace__(ns_pkg):
    """
    Uses namespace package to locate all namespaces with the ns_pkg as its root.

    For example in our system any namespace package that starts with volttron.services
    should be detected.

    NOTE: NO __init__.py file should ever be located within any package volttron.services or
            the importing will break

    @param: ns_pkg: Namespace to search for modules in.
    """
    # Specifying the second argument (prefix) to iter_modules makes the
    # returned name an absolute name instead of a relative one. This allows
    # import_module to work without having to do additional modification to
    # the name.
    return pkgutil.iter_modules(ns_pkg.__path__, ns_pkg.__name__ + ".")


def __discover_services__(namespace: str | ModuleType) -> List[str]:
    """
    Map all of the discovered namespaces to the volttron.services import.  Build
    a dictionary 'package' -> module.
    """
    if isinstance(namespace, str):
        namespace = importlib.import_module(namespace)

    # Add the namespace that is searched in case we need to load others in the future.
    __namespaces__.add(namespace.__name__)
    found: List[str] = []
    for finder, name, ispkg in __iter_namespace__(namespace):
        found.append(name)
        __discovered_plugins__[name] = importlib.import_module(name)
    return found


# """
# Manage the startup order of plugins available.  Note an error will
# be raised and the server will not startup if the plugin doesn't exist.
# The plugins that are within this same codebase hold the "default" services
# that should always be available in the system.  VOLTTRON requires that
# the services be started in a specific order for its processing to work as
# intended.
# """
# plugin_startup_order = [
#     "volttron.services.config_store",
#     "volttron.services.auth",
# ]
#
# plugin_disabled = ["volttron.services.health"]
#
# for p in plugin_startup_order:
#     if p not in __discovered_plugins__:
#         raise ValueError(f"Invalid plugin specified in plugin_startup_order {p}")
#     _log.info(f"Starting plugin: {p}, {__discovered_plugins__[p]}")
#
# for p, v in __discovered_plugins__.items():
#     if p not in plugin_startup_order and p not in plugin_disabled:
#         _log.info(f"Starting plugin {p}, {v}")
